from numbers import Number
from typing import Dict, List, Optional, Tuple, Union

import cimod
import numpy as np
import openjij

from jijmodeling.expression.expression import Expression

from jijzept.client import JijZeptClient
from jijzept.entity.schema import SolverType
from jijzept.post_api import post_instance_and_query
from jijzept.response import APIStatus, DimodResponse
from jijzept.sampler.jijmodel_post import JijModelingInterface


class JijSwapMovingSampler(JijModelingInterface):

    solver_type: SolverType = SolverType(queue_name="swapmovingsolver", solver="SA")
    jijmodeling_solver_type = SolverType(
        queue_name="swapmovingsolver", solver="SAParaSearch"
    )

    def __init__(
        self,
        token: str = None,
        url: Union[str, dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def get_result(self, solution_id: str):
        response = DimodResponse.empty_response(
            APIStatus.PENDING, self.client, solution_id
        )

        return response.get_result()

    @property
    def properties(self):
        return dict()

    @property
    def parameters(self):
        return dict()

    def _select_index_type_from_interactions(self, J: Optional[dict] = None):
        if J is None:
            return "IndexType.INT"
        elif isinstance(J, dict):
            if len(J) == 0:
                return "IndexType.INT"
            else:
                for key in J.keys():
                    if not isinstance(key, tuple):
                        raise TypeError("Invalid Interactions")
                    else:
                        for i in range(len(key)):
                            if isinstance(key[i], int):
                                return "IndexType.INT"
                            elif isinstance(key[i], str):
                                return "IndexType.STRING"
                            elif isinstance(key[i], tuple):
                                if len(key[i]) == 2:
                                    return "IndexType.INT_TUPLE_2"
                                elif len(key[i]) == 3:
                                    return "IndexType.INT_TUPLE_3"
                                elif len(key[i]) == 4:
                                    return "IndexType.INT_TUPLE_4"
                        raise TypeError("Invalid Interactions")
        else:
            raise TypeError("Invalid Interactions")

    def _select_index_type_from_linear(self, h: Optional[dict] = None):
        if h is None:
            return "IndexType.INT"
        elif isinstance(h, dict):
            if len(h) == 0:
                return "IndexType.INT"
            else:
                for key in h.keys():
                    if isinstance(key, int):
                        return "IndexType.INT"
                    elif isinstance(key, str):
                        return "IndexType.STRING"
                    elif isinstance(key, tuple) and len(key) == 2:
                        return "IndexType.INT_TUPLE_2"
                    elif isinstance(key, tuple) and len(key) == 3:
                        return "IndexType.INT_TUPLE_3"
                    elif isinstance(key, tuple) and len(key) == 4:
                        return "IndexType.INT_TUPLE_4"
                    raise TypeError("Invalid linear terms")
        else:
            raise TypeError("Invalid linear terms")

    def _to_serializable(
        self, offset, linear, interaction, penalties, constraints, vartype
    ):
        if len(linear) != 0:
            index_type = self._select_index_type_from_linear(linear)
        elif len(interaction) != 0:
            index_type = self._select_index_type_from_interactions(interaction)
        else:
            raise TypeError("Interactions are empty")

        if vartype == openjij.SPIN:
            vartype = "SPIN"
        elif vartype == openjij.BINARY:
            vartype = "BINARY"

        if constraints is None:
            constraints = []

        p_lambda = []
        p_key = []
        p_value = []
        p_constant = []

        if penalties is not None:
            for p in penalties:
                p_lambda.append(p[0])
                p_key.append(list(p[1].keys()))
                p_value.append(list(p[1].values()))
                p_constant.append(p[2])

        return {
            "data_format_type": "openjij.BinaryQuadraticModel",
            "model_type": "jijpro.BinaryModel",
            "index_type": index_type,
            "vartype": vartype,
            "offset": offset,
            "linear_key": list(linear.keys()),
            "linear_value": list(linear.values()),
            "interaction_key": list(interaction.keys()),
            "interaction_value": list(interaction.values()),
            "constraints": constraints,
            "penalties_lambda": p_lambda,
            "penalties_key": p_key,
            "penalties_value": p_value,
            "penalties_constant": p_constant,
        }

    def sample(
        self,
        bqm: Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel],
        constraints: Optional[List] = None,
        penalties: Optional[List] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        initial_state: Optional[Union[list, dict]] = None,
        updater: Optional[str] = None,
        sparse: Optional[bool] = None,
        reinitialize_state: Optional[bool] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """The simulated annealing is performed for BinaryQuadraticModel with the linear constraint conditions.
        Note here that the linear constraint conditions are strictly satisfied.

        Args:
            bqm (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel]): Binary quadratic model.
            constraints (Optional[List[List, int]]): The list of constraint conditions which can be represented as linear equations.
            penalties (Optional[List[float, List, int]]): The list of penarty terms which can be represented as linear equations.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            initial_state (Optional[dict], optional): Initial state. If `None`, this will be set automatically.
            updater (Optional[str], optional): Updater algorithm must be "swap moving" for now. If `None`, "swap moving" will be set.
            sparse (Optional[bool], optional): If `True`, only non-zero matrix elements are stored, which will save memory. If `None`, `True` will be set.
            reinitialize_state (Optional[bool], optional): If `True`, reinitialize state for each run. If `None`, `True` will be set.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        For `cimod.BinaryQuadraticModel` case:

        ```python
        import jijzept as jz
        import cimod
        bqm = cimod.BinaryQuadraticModel({0: -1, 2: -1}, {(0, 1): -1, (1, 3): -1}, "SPIN")
        sampler = jz.JijSwapMovingSampler(config='config.toml')
        response = sampler.sample(bqm, constraints=[([0,1], 0), ([2,3], 0)])
        ```

        """

        parameters = {
            "constraints": constraints,
            "penalties": penalties,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "num_sweeps": num_sweeps,
            "num_reads": num_reads,
            "initial_state": initial_state,
            "updater": updater,
            "sparse": sparse,
            "reinitialize_state": reinitialize_state,
            "seed": seed,
        }

        response = post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BQM",
            instance=self._to_serializable(
                bqm.get_offset(),
                bqm.get_linear(),
                bqm.get_quadratic(),
                penalties,
                constraints,
                bqm.vartype,
            ),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )
        return response

    def sample_ising(
        self,
        h: dict,
        J: dict,
        constraints: Optional[List] = None,
        penalties: Optional[List] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        initial_state: Optional[Union[list, dict]] = None,
        updater: Optional[str] = None,
        sparse: Optional[bool] = None,
        reinitialize_state: Optional[bool] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """The simulated annealing is performed for Ising models with the linear constraint conditions.
        Note here that the linear constraint conditions are strictly satisfied.

        Args:
            h (Dict): The linear terms.
            J (Dict): The quadratic terms.
            constraints (Optional[List[List, int]]): The list of constraint conditions which can be represented as linear equations.
            penalties (Optional[List[float, List, int]]): The list of penarty terms which can be represented as linear equations.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            initial_state (Optional[dict], optional): Initial state. If `None`, this will be set automatically.
            updater (Optional[str], optional): Updater algorithm must be "swap moving" for now. If `None`, "swap moving" will be set.
            sparse (Optional[bool], optional): If `True`, only non-zero matrix elements are stored, which will save memory. If `None`, `True` will be set.
            reinitialize_state (Optional[bool], optional): If `True`, reinitialize state for each run. If `None`, `True` will be set.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijSwapMovingSampler(config='config.toml')
        response = sampler.sample_ising(h={0: +0.5, 2: -0.5},
                                        J={(0,1):-1, (0,3):-1},
                                        constraints=[([0,1], 0), ([2,3], 0)])
        ```

        """

        parameters = {
            "constraints": constraints,
            "penalties": penalties,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "num_sweeps": num_sweeps,
            "num_reads": num_reads,
            "initial_state": initial_state,
            "updater": updater,
            "sparse": sparse,
            "reinitialize_state": reinitialize_state,
            "seed": seed,
        }

        response = post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BQM",
            instance=self._to_serializable(0.0, h, J, penalties, constraints, "SPIN"),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

        return response

    def sample_qubo(
        self,
        Q: dict,
        constraints: Optional[List] = None,
        penalties: Optional[List] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        initial_state: Optional[Union[list, dict]] = None,
        updater: Optional[str] = None,
        sparse: Optional[bool] = None,
        reinitialize_state: Optional[bool] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """The simulated annealing is performed for binary quadratic models with the linear constraint conditions.
        Note here that the linear constraint conditions are strictly satisfied.

        Args:
            Q (Dict): The linear or quadratic terms.
            constraints (Optional[List[List, int]]): The list of constraint conditions which can be represented as linear equations.
            penalties (Optional[List[float, List, int]]): The list of penarty terms which can be represented as linear equations.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            initial_state (Optional[dict], optional): Initial state. If `None`, this will be set automatically.
            updater (Optional[str], optional): Updater algorithm must be "swap moving" for now. If `None`, "swap moving" will be set.
            sparse (Optional[bool], optional): If `True`, only non-zero matrix elements are stored, which will save memory. If `None`, `True` will be set.
            reinitialize_state (Optional[bool], optional): If `True`, reinitialize state for each run. If `None`, `True` will be set.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijSwapMovingSampler(config='config.toml')
        response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1},
                                       constraints=[([0,1], 1), ([2,3], 1)])
        ```

        """

        parameters = {
            "constraints": constraints,
            "penalties": penalties,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "num_sweeps": num_sweeps,
            "num_reads": num_reads,
            "initial_state": initial_state,
            "updater": updater,
            "sparse": sparse,
            "reinitialize_state": reinitialize_state,
            "seed": seed,
        }

        response = post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BQM",
            instance=self._to_serializable(
                0.0, {}, Q, penalties, constraints, "BINARY"
            ),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

        return response

    def sample_hubo(
        self,
        J: dict,
        vartype: str,
        constraints: Optional[List] = None,
        penalties: Optional[List] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        initial_state: Optional[Union[list, dict]] = None,
        updater: Optional[str] = None,
        sparse: Optional[bool] = None,
        reinitialize_state: Optional[bool] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """The simulated annealing is performed for higher ordered unconstraind binary optimizations (hubo) with the linear constraint conditions.
        Note here that the linear constraint conditions are strictly satisfied.

        Args:
            J (Dict): Polynomial interactions.
            vartype (str): The variable type. "SPIN" or "BINARY".
            constraints (Optional[List[List, int]]): The list of constraint conditions which can be represented as linear equations.
            penalties (Optional[List[float, List, int]]): The list of penarty terms which can be represented as linear equations.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            initial_state (Optional[dict], optional): Initial state. If `None`, this will be set automatically.
            updater (Optional[str], optional): Updater algorithm must be "swap moving" for now. If `None`, "swap moving" will be set.
            sparse (Optional[bool], optional): Must be `True` for now. If `True`, only non-zero matrix elements are stored, which will save memory. If `None`, `True` will be set.
            reinitialize_state (Optional[bool], optional): If `True`, reinitialize state for each run. If `None`, `True` will be set.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijSwapMovingSampler(config='config.toml')
        response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN",
                                       constraints=[([0,1], 0), ([2,3,4], 1)])
        ```

        """

        parameters = {
            "constraints": constraints,
            "penalties": penalties,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "num_sweeps": num_sweeps,
            "num_reads": num_reads,
            "initial_state": initial_state,
            "updater": updater,
            "sparse": sparse,
            "reinitialize_state": reinitialize_state,
            "seed": seed,
        }

        response = post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BQM",
            instance=self._to_serializable(0.0, {}, J, penalties, constraints, vartype),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

        return response

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        updater: Optional[str] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of the simulated annealing.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            updater (Optional[str], optional): Updater algorithm must be "swap moving" for now. If `None`, "swap moving" will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        problem += jm.Constraint("one-hot", jm.Sum(i, x[i]) == 1)
        sampler = jz.JijSwapMovingSampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            beta_min=beta_min,
            beta_max=beta_max,
            num_sweeps=num_sweeps,
            num_reads=num_reads,
            updater=updater,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import jijzept.config.config_command
import jijzept.config.config_command as config_command
import jijzept.config.handle_config
import jijzept.config.handle_config as handle_config

from jijzept.config.handle_config import Config

__all__ = [
    "Config",
]



import json

from typing import Any, Dict, Optional

import requests, ujson, zstandard

from requests.exceptions import HTTPError

from jijzept.config import Config


class JijZeptClient:
    def __init__(
        self,
        url: Optional[str] = None,
        token: Optional[str] = None,
        proxy: Optional[str] = None,
        config: Optional[str] = None,
        config_env: str = "default",
    ):

        self.config = Config(url, token, proxy, config, config_env)
        # url = self.config.url
        # self.url = url if url[-1] != '/' else url[:-1]
        query_url = self.config.query_url
        self.query_url = query_url if query_url[-1] != "/" else query_url[:-1]

        post_url = self.config.post_url
        self.post_url = post_url if post_url[-1] != "/" else post_url[:-1]

        self.token = self.config.token

        self.proxy = self.config.proxy

        self.instance_id: Optional[str] = None
        self.req_solution_id: Optional[str] = None

    def post_instance(
        self, instance_type: str, instance: Dict[str, Any], endpoint: str = "/instance"
    ) -> Dict[str, str]:
        if not isinstance(instance_type, str):
            raise TypeError(
                "'instance_type' is `str`, not `{}`".format(type(instance_type))
            )
        if not isinstance(instance, dict):
            raise TypeError("instance is `dict`, not `{}`".format(type(instance)))

        endpoint = endpoint[:-1] if endpoint[-1] == "/" else endpoint

        # ------- Upload instance data to JijZept ---------
        upload_endpoint = self.post_url + endpoint + "/upload"
        headers = {
            "Content-Type": "application/zstd",
            "Ocp-Apim-Subscription-Key": self.token,
        }

        # encode instance
        json_data = ujson.dumps(instance)
        json_binary = json_data.encode("ascii")
        compressed_binary = zstandard.ZstdCompressor().compress(json_binary)
        res = requests.post(
            upload_endpoint,
            headers=headers,
            proxies=self.proxy,
            data=compressed_binary,
            stream=True,
        )

        status_check(res)

        # res_body => {
        #   "file_name": str
        # }
        res_body: dict = res.json()
        self.instance_data_id = res_body["instance_data_id"]
        # -----------------------------------------------

        # ---- `instance_type` registration to JijZept-API -----
        regist_endpoint = self.post_url + endpoint
        req_body = ujson.dumps(
            {"instance_type": instance_type, "instance_data_id": self.instance_data_id}
        )
        headers = {
            "Content-Type": "application/json",
            "Ocp-Apim-Subscription-Key": self.token,
        }

        res = requests.post(
            regist_endpoint, headers=headers, proxies=self.proxy, data=req_body
        )
        status_check(res)

        self.instance_id = res.json()["instance_id"]
        # ------------------------------------------------------
        return self.instance_id

    def submit_solve_query(
        self,
        queue_name: str,
        solver_name: str,
        parameters: dict,
        instance_id: Optional[str] = None,
        timeout: Optional[float] = None,
        endpoint: str = "/query/solution",
    ):
        if self.instance_id is None and instance_id is None:
            message = "solve_request() missing 1 "
            message += "require positional argument: 'instance_id'"
            raise TypeError(message)

        if not isinstance(solver_name, str):
            raise TypeError("`solver_name` is str. not {}".format(type(solver_name)))

        if not isinstance(parameters, dict):
            raise TypeError("`parameters` is dict. not {}".format(type(parameters)))

        instance_id = self.instance_id if instance_id is None else instance_id

        query_endopoint = self.query_url + endpoint
        headers = {
            "Content-Type": "application/json",
            "Ocp-Apim-Subscription-Key": self.token,
        }

        json_data = ujson.dumps(
            {
                "instance_id": instance_id,
                "solver_params": parameters,
                "queue_name": queue_name,
                "solver": solver_name,
                "timeout": timeout,
            }
        )
        res = requests.post(
            query_endopoint, headers=headers, proxies=self.proxy, data=json_data
        )
        status_check(res)

        # res_body => {
        #   "solution_id": str,
        # }
        res_body: dict = res.json()
        self.req_solution_id = res_body["solution_id"]

        return res_body

    def fetch_result(
        self, solution_id: Optional[str] = None, endpoint: str = "/query/solution"
    ) -> dict:
        if self.req_solution_id is None and solution_id is None:
            message = "fetch_result() missing 1 "
            message += "require positional argument: 'solution_id'"
            raise TypeError(message)

        solution_id = self.req_solution_id if solution_id is None else solution_id

        fetch_endpoint = self.query_url + endpoint

        headers = {"Ocp-Apim-Subscription-Key": self.token}

        params = {"solution_id": solution_id}

        res = requests.get(
            fetch_endpoint, headers=headers, proxies=self.proxy, params=params
        )

        status_check(res)

        res_body = res.json()
        if res_body is None:
            raise HTTPError("solution_id is not found")

        return res_body


def status_check(res) -> None:
    try:
        res.raise_for_status()
    except requests.exceptions.HTTPError as e:
        res = e.response
        try:
            res_body = res.json()
        except json.decoder.JSONDecodeError:
            raise requests.exceptions.HTTPError(e)

        raise requests.exceptions.HTTPError(e, res_body)



import os

from pathlib import Path
from typing import Dict, Optional, TypeVar, Union
from urllib.parse import ParseResult, urlparse

import toml

_T0 = TypeVar("_T0")

# DEFAULT SETTING
CONFIG_PATH = str(Path.home()) + "/.jijzept/"
HOST_URL = ""
DEFAULT_CONFIG_FILE = "config.toml"


def create_config(token: str, host_url=HOST_URL, config_path=CONFIG_PATH):
    Path(config_path).mkdir(parents=True, exist_ok=True)

    config_dict = {"default": {"url": host_url, "token": token}}
    # save config file's
    config_path = config_path if config_path[-1] == "/" else config_path + "/"
    config_file_name = config_path + DEFAULT_CONFIG_FILE
    with open(config_file_name, mode="w") as f:
        toml.dump(config_dict, f)

    return config_file_name


def load_config(file_name: str, config="default") -> dict:
    """load config file (TOML file)

    Args:
        file_name (str): path to config file.
        config (str, optional): loading enviroment name. Defaults to 'default'.

    Raises:
        TypeError: if 'config' enviroment is not defined in config file.

    Returns:
        dict: {'token': 'xxxx', 'url': 'xxxx'}
    """
    p_rel = Path(file_name)
    with open(p_rel) as f:
        toml_setting_file = toml.load(f)
    if config not in toml_setting_file:
        raise TypeError(
            "'{}' is not define in config file ({}).".format(config, file_name)
        )
    return toml_setting_file


def _query_and_post_url(url: _T0) -> Dict[str, Union[str, _T0]]:
    parsed = urlparse(url)
    if not parsed.path:
        path = parsed.path + '/query'  
    elif parsed.path[-1] != '/':
        path = parsed.path + '/query'
    else: 
        path = parsed.path + 'query'
    parsed_post_url = ParseResult(
        parsed.scheme,
        parsed.netloc,
        path,
        parsed.params,
        parsed.query,
        parsed.fragment,
    ).geturl()
    return {"query_url": url, "post_url": parsed_post_url}


class Config:
    """JijZept API Config

    Attributes:
        query_url (str): Endpoint for Query API.
        post_url (str): Endpoint for Post Instance API.
        token (str): Secret token to connect API.
    """

    def __init__(
        self,
        url: Optional[str] = None,
        token: Optional[str] = None,
        proxy: Optional[str] = None,
        config: Optional[str] = None,
        config_env: str = "default",
    ):

        # setting Proxy info
        if proxy is not None:
            if not isinstance(proxy, str):
                raise TypeError("'proxy' is `str`.")
            self.proxy = proxy

        if url is not None and token is not None:
            if not isinstance(url, str) or not isinstance(token, str):
                raise TypeError("'url' and 'token' are `str`.")
            endpoints = _query_and_post_url(url)
            self.post_url = endpoints["post_url"]
            self.query_url = endpoints["query_url"]
            self.token = token

        else:
            if config is None:
                config_file = CONFIG_PATH + DEFAULT_CONFIG_FILE
            else:
                config_file = config
            if not os.path.exists(config_file):
                message = "A configuration file does not exist.\n"
                message += "set arguments of __init__(): 'url' and 'token'\n"
                message += "or create config file"
                message += " using `jijzept create` command."
                raise ValueError(message)
            _config = load_config(config_file, config_env)
            self.token = _config[config_env]["token"]

            # Proxy Setup
            if "proxy" in _config[config_env]:
                self.proxy = {
                    "http": _config[config_env]["proxy"],
                    "https": _config[config_env]["proxy"],
                }
            else:
                self.proxy = None

            # If the query_url and post_url are set separately,
            # they will be set as API endpoints in preference to each other.
            query_url = "query_url"
            post_url = "post_url"
            query_url_value = None
            post_url_value = None

            if "url" in _config[config_env]:
                endpoints = _query_and_post_url(_config[config_env]["url"])
                query_url_value = endpoints[query_url]
                post_url_value = endpoints[post_url]
            if query_url in _config[config_env]:
                query_url_value = _config[config_env][query_url]
            if post_url in _config[config_env]:
                post_url_value = _config[config_env][post_url]

            if query_url_value is None or post_url_value is None:
                raise ValueError(
                    "The `url` parameter is required"
                    + " for the K environment in the configuration file."
                )

            self.query_url = query_url_value
            self.post_url = post_url_value



from setuptools import setup

setup(
  install_requires =[
    "jijmodeling <=  0.9.23",
    "openjij >=0.5.0, <0.6.0",
    "jij-cimod >= 1.4.1, <1.5.0",
    'dimod >=0.9.14, < 0.11.0; python_version < "3.10"',
    'dimod >= 0.9.1, < 0.12.0; python_version >= "3.10"',
    "numpy<1.22",
    "requests",
    "toml",
    "zstandard",
    "pydantic",
    "ujson",
    "click",
  ]
)



from numbers import Number
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np

from jijmodeling.expression.expression import Expression

from jijzept.entity.schema import SolverType
from jijzept.sampler.jijmodel_post import JijModelingInterface
from jijzept.sampler.sampler import JijZeptSampler


class JijSBMSampler(JijZeptSampler, JijModelingInterface):
    solver_type = SolverType(queue_name="sbmsolver", solver="SBM")
    jijmodeling_solver_type = SolverType(queue_name="sbmsolver", solver="SBMParaSearch")

    def sample(
        self,
        bqm,
        steps=None,
        loops=None,
        maxwait=None,
        target=None,
        prefer="auto",
        stats="none",
        dt=None,
        C=None,
        timeout=None,
        sync=True,
        queue_name: Optional[str] = None,
    ) -> Any:
        """Samples by SBM.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`): Binary quadratic model.
            steps (int, optional): The number of steps in SBM calculation. If `steps` is set to `0`, then the number of steps is set automatically. Defaults to 0.
            loops (int, optional): The number of loops in SBM calculation. If `loops` is set to `0`, then calculation will be repeated to the maximum calculation time. Defaults to 1.
            maxwait (int, optional): The maximum waiting time in seconds. Defaults to None.
            target (float, optional): The end condition of calculation. Defaults to None.
            prefer (str, optional): Select `'speed'` or `'auto'`. Defaults to 'auto'.
            stats (str, optional): Select `'none'`, `'summary'`, or `'full`. Defaults to 'none'.
            dt (float, optional): Time step width. Defaults to 0.1.
            C (float, optional): Positive constant coefficient. If `0` is set to `C`, then the value of `C` is set automatically. Defaults to 0.
            timeout (float, optional): The maximum calculation time in seconds. Defaults to None.
            sync (bool, optional): Synchronization mode. Defaults to True.
            queue_name (str, optional): queue_name.

        Returns:
            :obj:`dimod.SampleSet`: Stores minimum energy sample.

        """
        return super().sample(
            bqm,
            steps=steps,
            loops=loops,
            maxwait=maxwait,
            target=target,
            prefer=prefer,
            stats=stats,
            dt=dt,
            C=C,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        steps=None,
        loops=None,
        maxwait=None,
        target=None,
        prefer="auto",
        stats="none",
        dt=None,
        C=None,
        timeout: Optional[float] = None,
        sync=True,
        queue_name: Optional[str] = None,
    ):
        """sample_model.

        Args:
            model (Expression): model
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): feed_dict
            multipliers (Dict[str, Number]): multipliers
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool): search
            steps (int, optional): The number of steps in SBM calculation. If `steps` is set to `0`, then the number of steps is set automatically. Defaults to 0.
            loops (int, optional): The number of loops in SBM calculation. If `loops` is set to `0`, then calculation will be repeated to the maximum calculation time. Defaults to 1.
            maxwait (int, optional): The maximum waiting time in seconds. Defaults to None.
            target (float, optional): The end condition of calculation. Defaults to None.
            prefer (str, optional): Select `'speed'` or `'auto'`. Defaults to 'auto'.
            stats (str, optional): Select `'none'`, `'summary'`, or `'full`. Defaults to 'none'.
            dt (float, optional): Time step width. Defaults to 0.1.
            C (float, optional): Positive constant coefficient. If `0` is set to `C`, then the value of `C` is set automatically. Defaults to 0.
            timeout (float, optional): The maximum calculation time in seconds. Defaults to None.
            sync (bool, optional): Synchronization mode. Defaults to True.
            queue_name (str, optional): queue_name.
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            steps=steps,
            loops=loops,
            maxwait=maxwait,
            target=target,
            prefer=prefer,
            stats=stats,
            dt=dt,
            C=C,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )



from numbers import Number
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from jijmodeling.expression.expression import Expression

from jijzept.entity.schema import SolverType
from jijzept.response import DimodResponse
from jijzept.sampler.jijmodel_post import JijModelingInterface
from jijzept.sampler.sampler import JijZeptSampler


class JijDWaveLeapSampler(JijZeptSampler, JijModelingInterface):
    solver_type = SolverType(queue_name="dwaveleapsolver", solver="DWave")
    jijmodeling_solver_type = SolverType(
        queue_name="dwaveleapsolver", solver="DWaveParaSearch"
    )

    def sample(
        self,
        bqm,
        num_reads: Optional[int] = None,
        annealing_time: Optional[float] = None,
        auto_scale: Optional[bool] = None,
        annealing_schedule: Optional[List[List[float]]] = None,
        timeout: Optional[float] = None,
        sync=True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Samples by D-Wave Sampler

        Args:
            bqm (dimod.BinaryQuadraticModel): Binary quadratic model.
            num_reads (Optional[int], optional): number of annealing sample. Defaults to 1.
            annealing_time (Optional[float], optional): quantum annealing time [μs]. Defaults to None.
            auto_scale (Optional[bool], optional): auto scale strength of interaction. Defaults to False.
            timeout (Optional[float], optional): timeout
            sync (bool): set sync mode. Defaults to True.
            queue_name (str, optional): queue_name.

        Returns:
            :obj:`dimod.SampleSet`: Stores minimum energy samples.
        """

        return super().sample(
            bqm,
            num_reads=num_reads,
            annealing_time=annealing_time,
            auto_scale=auto_scale,
            annealing_schedule=annealing_schedule,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]] = {},
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        num_reads: Optional[int] = None,
        annealing_time: Optional[float] = None,
        auto_scale: Optional[bool] = None,
        annealing_schedule: Optional[List[List[float]]] = None,
        timeout: Optional[float] = None,
        sync=True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """sample using jijmodeling

        Args:
            model (Expression): model
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): feed_dict
            multipliers (Dict[str, Number]): multipliers
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool): search
            num_search (int): num_search
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            num_reads (Optional[int]): num_reads
            annealing_time (Optional[float]): annealing_time
            auto_scale (Optional[bool]): auto_scale
            annealing_schedule (Optional[List[List[float]]]): annealing_schedule
            timeout (Optional[float]): timeout
            sync (bool): synchronous mode
            queue_name (str, optional): queue_name.

        Returns:
            DimodResponse:
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            num_reads=num_reads,
            annealing_time=annealing_time,
            auto_scale=auto_scale,
            annealing_schedule=annealing_schedule,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )
